Imports System.Drawing.Drawing2D
''' <summary>
''' Basisklasse fr selbst gezeichnete scrollbare Listen-Controls 
''' - auch fr eine groe Zahl Items -
''' Jedes Item liefert bei Bedarf ein Objekt (ListPainterItem), das dieses Item zeichnet.
''' Nur die Objekte, die wirklich sichtbar sind, bleiben im Arbeitsspeicher
''' </summary>

Public MustInherit Class ListPaintControlBase
    Inherits Control

#Region " Private Felder "

    ' ScrollBar
    Private WithEvents m_ScrollBar As GlassScrollBar
    ' zuletzt beim manuellen Scrollen eingestellter Wert,
    ' sorgt bei Hhennderung des Controls dafr, dass das TopItem auch oben bleibt
    Private m_LastScrolledValue As Integer = 0
    ' Flag zum Verzichten auf ScrollBar-Events bei internem Scrollen
    Private m_ScrollBarEventsEnabled As Boolean = True


    ' Liste, die die angezeigten Items enthlt, ein SelectedItem verwaltet
    ' und bei Bedarf Changed-Events liefert
    Private WithEvents m_Items As AddRemoveSelectionList(Of ListItem)
    ' Intern bentigte Kopie der Liste mit gerade sichtbaren Items
    Private m_VisibleItems As New List(Of ListItem)
    ' Dictionary mit je einem Item und zugehrigem zeichenbarem Objekt
    Private m_KnownItems As New Dictionary(Of ListItem, ListPaintItem)


    ' Interne Hhenangabe
    Private m_Height As Integer = 0

    ' Hhe eines Items
    Private m_ItemHeight As Integer = 56
    ' angezeigte Breite eines Items
    Private m_ItemWidth As Integer = 0

    ' Selektiertes Item
    Private m_SelectedItem As ListItem
    ' Gehoovertes Item
    Private m_HooveredItem As ListItem
    ' Item ber dem der linke MouseButton gedrckt wurde 
    Private m_MouseDownItem As ListItem


    ' Renderer zum einfachen ndern der Farben des Controls
    Private WithEvents m_Renderer As ColorWheelToolStripRenderer

    ' LinearGradientBrushes und zugehrige Rechtecke zum Zeichnen 
    ' der Hintergrund-Flchen
    Private m_LeftGradientBrush As LinearGradientBrush
    Private m_RightGradientBrush As LinearGradientBrush
    Private m_LeftGradientRect As RectangleF
    Private m_RightGradientRect As RectangleF

#End Region

#Region " Konstruktoren "

    ''' <summary>
    ''' Konstruktor mit einer vordefinierten Liste, die von AddRemoveSelectionList erbt
    ''' </summary>
    ''' <param name="Items">von AddRemoveSelectionList abgeleitete Listen-Instanz</param>
    Public Sub New(ByVal Items As AddRemoveSelectionList(Of ListItem))
        ' Listen-Instanz bernehmen
        m_Items = Items
        ' Renderer mit verwendeten Farben und Pens erstellen
        m_Renderer = New ColorWheelToolStripRenderer(New ColorWheelProfessionalTable)

        ' ScrollBar instanziieren
        m_ScrollBar = New GlassScrollBar
        ' ScrollBar erhlt eigenen Renderer als Referenz
        m_ScrollBar.Renderer = m_Renderer
        ' Renderer-Basisfarbe auf blau einstellen
        m_Renderer.ControlHue = 0.6R

        ' SmallChange der ScrollBar einstellen
        m_ScrollBar.SmallChange = 14
        ' ScrollBar zur eigenen Controls-Auflistung hinzufgen
        Me.Controls.Add(m_ScrollBar)

        ' DoubleBuffering aktivieren
        Me.SetStyle(ControlStyles.DoubleBuffer Or ControlStyles.UserPaint _
        Or ControlStyles.AllPaintingInWmPaint, True)

        ' gegebenenfalls enthlt die im Konstruktor bergebene Liste schon Elemente
        ' - dann also die Elemente auch anzeigen
        OnItemsChanged()
    End Sub

#End Region

#Region " Properties "

    ''' <summary>
    ''' gibt einen ColorWheelToolStripRenderer zur Definition der Farben dieses Controls
    '''  zurck oder legt diesen fest
    ''' </summary>
    Property Renderer() As ColorWheelToolStripRenderer
        Get
            Return m_Renderer
        End Get
        Set(ByVal value As ColorWheelToolStripRenderer)
            ' nur wenn tatschlich ein Renderer angegeben wurde
            If value IsNot Nothing Then
                m_Renderer = value
                ' Hintergrund-Gradient-Brushes neu definieren
                GetBackgroundBrushes()
                ' Falls vorhanden, die Farben der angezeigten Items neu festlegen
                RefreshItemColors()
                ' auch die ScrollBar verwendet den Renderer
                m_ScrollBar.Renderer = value
                ' Control zum Neu-Zeichnen zwingen
                Invalidate()
            End If
        End Set
    End Property


    ''' <summary>
    ''' gibt die Liste mit anzuzeigenden Elementen zurck oder legt diese fest
    ''' </summary>
    Property Items() As AddRemoveSelectionList(Of ListItem)
        Get
            Return m_Items
        End Get
        Set(ByVal value As AddRemoveSelectionList(Of ListItem))
            ' nur wenn es sich tatschlich um eine neue Liste handelt
            If m_Items IsNot value AndAlso value IsNot Nothing Then
                ' Referenz zur Liste bernehmen
                m_Items = value
                ' auf Events der ScrollBar vorbergehend verzichten
                m_ScrollBarEventsEnabled = False
                ' Value-Eigenschaft der ScrollBar zurck setzen
                m_LastScrolledValue = 0
                m_ScrollBar.Value = 0
                ' auf Events der ScrollBar wieder zulassen
                m_ScrollBarEventsEnabled = True
                ' Elemente der Liste neu zeichnen
                OnItemsChanged()
            End If
        End Set
    End Property

#End Region

#Region " Methoden "

    ''' <summary>
    ''' der Renderer liefert dieses Event bei nderung der Farbauswahl
    ''' </summary>
    Private Sub m_Renderer_Changed() Handles m_Renderer.Changed
        ' Hintergrund-Gradient-Brushes neu definieren
        GetBackgroundBrushes()
        ' Falls vorhanden, die Farben der angezeigten Items neu festlegen
        RefreshItemColors()
        ' Control zum Neu-Zeichnen zwingen
        Invalidate()
    End Sub

    ''' <summary>
    ''' legt die Farben aller angezeigten Items nach nderung im Renderer neu fest
    ''' </summary>
    Private Sub RefreshItemColors()
        For Each Item As ListItem In m_KnownItems.Keys
            m_KnownItems(Item).RefreshColors()
        Next
    End Sub


    ''' <summary>
    ''' ermittelt nach Grennderung des Controls oder Ein- bzw. Ausblenden der ScrollBar 
    ''' die LinearGradientBrushes und dazu gehrende Rechtecke 
    ''' (fr das Zeichnen des Control-Hintergrundes) neu
    ''' </summary>
    Private Sub GetBackgroundBrushes()
        ' Breite einer Gradient-Flche (muss grer als 0 sein)
        Dim GradientWidth As Integer
        If m_ScrollBar.Visible Then
            GradientWidth = Math.Min(12, Math.Max(2, (Me.Width - m_ScrollBar.Width) \ 2))
        Else
            GradientWidth = Math.Min(12, Math.Max(2, Me.Width \ 2))
        End If

        ' Rechteck fr linken Gradienten festlegen
        m_LeftGradientRect = New RectangleF(0, 0, GradientWidth, Math.Max(Me.Height, 1))
        ' Rechteck fr rechten Gradienten festlegen
        If m_ScrollBar.Visible Then
            m_RightGradientRect = New RectangleF(Me.Width - GradientWidth - m_ScrollBar.Width, 0, GradientWidth, Math.Max(Me.Height, 1))
        Else
            m_RightGradientRect = New RectangleF(Me.Width - GradientWidth, 0, GradientWidth, Math.Max(Me.Height, 1))
        End If

        Dim RightBrushRect As New RectangleF(m_RightGradientRect.X - 1.0F, m_RightGradientRect.Y, m_RightGradientRect.Width, m_RightGradientRect.Height)

        ' Brushes festlegen
        m_LeftGradientBrush = New LinearGradientBrush(m_LeftGradientRect, m_Renderer.ColorTable.ToolStripGradientBegin, m_Renderer.ColorTable.ToolStripGradientMiddle, LinearGradientMode.Horizontal)
        m_RightGradientBrush = New LinearGradientBrush(RightBrushRect, m_Renderer.ColorTable.ToolStripGradientMiddle, m_Renderer.ColorTable.ToolStripGradientEnd, LinearGradientMode.Horizontal)
    End Sub

    ''' <summary>
    ''' legt nach nderung des SelectedItem in der aktuellen Item-Liste oder Wechsel der
    ''' Liste selbst zugehrige PaintItems als selektiert oder deselektiert fest
    ''' </summary>
    ''' <param name="Item">Neu selektiertes Item</param>
    Private Sub SetSelected(ByVal Item As ListItem)
        ' nur wenn sich SelectedItem tatschlich ndert
        If m_SelectedItem IsNot Item Then

            ' altes PaintItem wird deselektiert
            If m_SelectedItem IsNot Nothing Then
                If m_KnownItems.ContainsKey(m_SelectedItem) Then
                    m_KnownItems(m_SelectedItem).Selected = False
                End If
            End If

            ' SelectedItem bernehmen
            m_SelectedItem = Item

            ' neues PaintItem wird selektiert
            If m_SelectedItem IsNot Nothing Then
                If m_KnownItems.ContainsKey(m_SelectedItem) Then
                    m_KnownItems(m_SelectedItem).Selected = True
                End If
            End If

        End If
    End Sub

    ''' <summary>
    ''' legt nach Hoovern mit der Maus das gehooverte Item neu fest
    ''' </summary>
    ''' <param name="Item">Neu gehoovertes Item</param>
    Private Sub SetHoovered(ByVal Item As ListItem)
        ' nur wenn sich HooveredItem tatschlich ndert
        If m_HooveredItem IsNot Item Then

            ' altes PaintItem wird unhoovered
            If m_HooveredItem IsNot Nothing Then
                If m_KnownItems.ContainsKey(m_HooveredItem) Then
                    m_KnownItems(m_HooveredItem).Hoovered = False
                End If
            End If

            ' HooveredItem bernehmen
            m_HooveredItem = Item

            ' neues PaintItem wird gehoovert
            If m_HooveredItem IsNot Nothing Then
                If m_KnownItems.ContainsKey(m_HooveredItem) Then
                    m_KnownItems(m_HooveredItem).Hoovered = True
                End If
            End If
        End If
    End Sub

    ''' <summary>
    ''' legt nach MouseDown das MouseDown-Item neu fest
    ''' </summary>
    ''' <param name="Item">Neues MouseDown-Item</param>
    Private Sub SetMouseDown(ByVal Item As ListItem)
        ' nur wenn sich MouseDown-Item tatschlich ndert
        If m_MouseDownItem IsNot Item Then

            ' altes PaintItem verliert den MouseDown-Status
            If m_MouseDownItem IsNot Nothing Then
                If m_KnownItems.ContainsKey(m_MouseDownItem) Then
                    m_KnownItems(m_MouseDownItem).MouseDown = False
                End If
            End If

            ' MouseDown-Item bernehmen
            m_MouseDownItem = Item

            ' neues PaintItem erhlt den MouseDown-Status
            If m_MouseDownItem IsNot Nothing Then
                If m_KnownItems.ContainsKey(m_MouseDownItem) Then
                    m_KnownItems(m_MouseDownItem).MouseDown = True
                End If
            End If
        End If
    End Sub


    ''' <summary>
    ''' Handler fr das Changed-Event der Item-Liste
    ''' </summary>
    Private Sub m_Items_Changed() Handles m_Items.Changed
        ' angezeigte Items neu ermitteln und gegebenenfalls neu zeichnen
        OnItemsChanged()
    End Sub

    ''' <summary>
    ''' berechnet die sichtbaren Items neu und aktualisiert bei Notwendigkeit die ScrollBar
    ''' und sorgt bei Notwendigkeit fr neues Zeichnen des Controls
    ''' </summary>
    Private Sub OnItemsChanged()

        ' vorbergehend auf Events von der ScrollBar verzichten
        m_ScrollBarEventsEnabled = False

        ' Hhe der Liste in Pixeln
        Dim ListHeight As Integer = m_Items.Count * m_ItemHeight

        If ListHeight > m_Height Then
            ' Wenn die Liste hher als das Control ist, muss die ScrollBar angezeigt werden
            m_ScrollBar.Visible = True
            ' Breite eines Items festlegen
            m_ItemWidth = Me.Width - m_ScrollBar.Width - 2
            ' Maximum der ScrollBar festlegen
            m_ScrollBar.Maximum = ListHeight - 1
            ' maximal mit der ScrollBar beim manuellen scrollen erreichbarer Wert
            Dim MaxScrollableValue As Integer = ListHeight - m_Height

            ' neuer Wert, den die ScrollBar erhlt, wenn SelectedItem nicht bercksichtigt wird
            Dim NewScrollValue As Integer = Math.Max(Math.Min(m_ScrollBar.Value, ListHeight - Me.Height), 0)
            If m_Items.SelectedItem IsNot Nothing Then
                ' wenn ein SelectedItem existiert,
                ' sicher stellen, dass dieses vollstndig sichtbar ist
                Dim Index As Integer = m_Items.IndexOf(m_Items.SelectedItem)
                Dim TopPos As Integer = Index * m_ItemHeight
                If TopPos < NewScrollValue Then
                    NewScrollValue = TopPos
                ElseIf TopPos + m_ItemHeight > NewScrollValue + m_Height Then
                    NewScrollValue = Math.Max(TopPos + m_ItemHeight - m_Height, 0)
                End If
            End If

            ' neuen Wert fr die ScrollBar bernehmen
            m_ScrollBar.Value = Math.Min(Math.Max(NewScrollValue, 0), MaxScrollableValue)
            ' LargeChange der ScrollBar entspricht der Hhe des Controls
            m_ScrollBar.LargeChange = m_Height
        Else
            ' Wenn die Liste kleiner als das Control ist, wird auf die ScrollBar verzichtet
            m_ScrollBar.Visible = False
            ' Breite eines Items festlegen
            m_ItemWidth = Me.Width - 1
            ' Value-Eigenschaft der ScrollBar ist 0
            m_ScrollBar.Value = 0
        End If

        ' notwendige PaintItems neu ermittlen
        ReSetVisibleItems()
        ' Falls notwendig, neu selektiertes Item als selektiert markieren
        ' und ehemals selektiertes Item deselektieren
        SetSelected(m_Items.SelectedItem)

        ' Events der ScrollBar wieder zulassen
        m_ScrollBarEventsEnabled = True

        ' zuletzt eingestellten Wert der ScrollBar merken
        m_LastScrolledValue = m_ScrollBar.Value

        ' Control neu zeichnen
        Invalidate()
    End Sub


    ''' <summary>
    ''' ermittelt notwendige PaintItems neu und entfernt berflssige Referenzen aus dem 
    ''' PaintItem-Dictionary und sorgt fr Aktualisierung der Anzeigebreite der PaintItems
    ''' </summary>
    Private Sub ReSetVisibleItems()

        ' Liste mit sichtbaren Items fr aktuellen ScrollBar-Wert erstellen
        Dim NewItems As List(Of ListItem) = GetVisibleItems()

        ' Flag mit Information, ob eine Aktuelaisierung der PaintItem-Dictionary-Elemente notwendig ist
        Dim ItemsChanged As Boolean = False

        If m_VisibleItems.Count <> NewItems.Count Then
            ' wenn sich die Anzahl sichtbarer Items ndert, muss das PaintItem-Dictionary auf jeden Fall 
            ' aktualisiert werden
            ItemsChanged = True
        ElseIf m_VisibleItems.Count > 0 Then
            ' wenn sich die Anzahl sichtbarer Items gleich bleibt, muss das PaintItem-Dictionary nur dann
            ' aktualisiert werden, wenn einzelne Items gendert sind
            For I As Integer = 0 To m_VisibleItems.Count - 1
                If m_VisibleItems(I) IsNot NewItems(I) Then
                    ItemsChanged = True
                    Exit For
                End If
            Next
        End If


        If ItemsChanged Then
            ' nur wenn nderung erforderlich

            ' Auflistung mit neu sichtbaren Items merken
            m_VisibleItems = NewItems

            ' Auflistung mit aus dem PaintItem-Dictionary zu entfernenden Items
            Dim ItemsToRemove As New List(Of ListItem)

            ' jeder Key im PaintItem-Dictionary, der nicht in der neuen Liste mit sichtbaren Items 
            ' enthalten sind, wird der ItemsToRemove-Auflistung hinzugefgt
            For Each Item As ListItem In m_KnownItems.Keys
                If Not m_VisibleItems.Contains(Item) Then
                    ItemsToRemove.Add(Item)
                End If
            Next

            ' fr alle Items in der neuen Liste mit sichtbaren Items, fr die kein Key im PaintItem-Dictionary
            ' enthalten ist, wird eine neues PaintItem-Objekt erstellt, und dieses Objekt mit Item als Key
            ' dem Dictionary hinzugefgt
            For Each Item As ListItem In m_VisibleItems
                If Not m_KnownItems.ContainsKey(Item) Then
                    Dim NewPaintItem As ListPaintItem = Item.CreatePaintItem(m_Items.SelectedItem Is Item)
                    NewPaintItem.Control = Me
                    m_KnownItems.Add(Item, NewPaintItem)
                End If
            Next

            ' Im Anschluss alle fr in der Lschliste enthaltenen Elemente den entsprechenden 
            ' Eintrag im Dictionary entfernen und gegebenenfalls im PaintItem verwendete Ressourcen 
            ' frei geben
            For Each Item As ListItem In ItemsToRemove
                m_KnownItems(Item).DisposeBitmap()
                m_KnownItems.Remove(Item)
            Next
        End If

        ' fr alle PaintItems das Anzeigerechteck aktualisieren
        If m_VisibleItems.Count > 0 Then
            Dim TopItem As ListItem = m_VisibleItems(0)
            Dim TopPosition As Integer = m_Items.IndexOf(TopItem) * m_ItemHeight - m_ScrollBar.Value

            For I As Integer = 0 To m_VisibleItems.Count - 1
                m_KnownItems(m_VisibleItems(I)).Rectangle = New Rectangle(0, TopPosition, m_ItemWidth, m_ItemHeight)
                TopPosition += m_ItemHeight
            Next
        End If

    End Sub



    ''' <summary>
    ''' liefert eine Liste mit aktuell im Control sichtbaren Items
    ''' </summary>
    Private Function GetVisibleItems() As List(Of ListItem)
        ' neue leere Auflistung erstellen
        Dim Result As New List(Of ListItem)

        ' nur wenn die Item-Auflistung wirklich Elemente enthlt
        If m_Items.Count > 0 Then
            ' welche Items sichtbar sind, hngt von der Value-Eigenschaft der ScrollBar,
            ' der Hhe des Controls und der Hhe eines einzelnen Items ab
            Dim FirstIndex As Integer = m_ScrollBar.Value \ m_ItemHeight
            Dim LastIndex As Integer = Math.Min((m_ScrollBar.Value + m_Height) \ m_ItemHeight, m_Items.Count - 1)
            Dim Length As Integer = LastIndex - FirstIndex + 1

            Result.AddRange(m_Items.GetRange(FirstIndex, Length))
        End If

        ' Auflistung zurck geben
        Return Result
    End Function


    ''' <summary>
    ''' Handler fr das Scroll-Event der ScrollBar
    ''' </summary>
    Private Sub m_ScrollBar_Scroll(ByVal sender As Object, ByVal e As System.Windows.Forms.ScrollEventArgs) Handles m_ScrollBar.Scroll
        ' wenn auf ScrollBar-Events verzichtet werden soll, nichts tun
        If Not m_ScrollBarEventsEnabled Then Exit Sub

        ' anzuzeigende PaintItems neu ermitteln
        ReSetVisibleItems()
        ' aktuellen Wert der ScrollBar merken
        m_LastScrolledValue = m_ScrollBar.Value
        ' Control neu zeichnen
        Invalidate()

    End Sub

    ''' <summary>
    ''' Handler fr das MouseDown-Event des Controls
    ''' </summary>
    Private Sub ListPaintControl_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        ' Item ermitteln, ber dem die Maus gedrckt wurde
        Dim NewMouseDown As ListItem = Nothing
        For Each Item As ListItem In m_KnownItems.Keys
            If m_KnownItems(Item).Rectangle.Contains(e.Location) Then
                NewMouseDown = Item
                Exit For
            End If
        Next
        ' ausgewhltes Item als MouseDownItem merken und gegebenenfalls die Anzeige des zugehrigen PaintItems aktualisieren
        SetMouseDown(NewMouseDown)
    End Sub

    ''' <summary>
    ''' Handler fr das MouseLeave-Event des Controls
    ''' </summary>
    Private Sub ListPaintControl_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseLeave
        ' wenn das Control den Fokus verliert, verliert das aktuell gehooverte Item seinen Hoovered-Status
        SetHoovered(Nothing)
    End Sub

    ''' <summary>
    ''' Handler fr das MouseMove-Event des Controls
    ''' </summary>
    Private Sub ListPaintControl_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove
        ' nur wenn kein Button gedrckt ist
        If e.Button <> Windows.Forms.MouseButtons.None Then Exit Sub

        ' Item ermitteln, ber dem die Maus sich befindet
        Dim NewHoovered As ListItem = Nothing
        For Each Item As ListItem In m_KnownItems.Keys
            If m_KnownItems(Item).Rectangle.Contains(e.Location) Then
                NewHoovered = Item
                Exit For
            End If
        Next

        ' ausgewhltes Item als HooveredItem merken und gegebenenfalls die Anzeige des zugehrigen PaintItems aktualisieren
        SetHoovered(NewHoovered)
    End Sub

    ''' <summary>
    ''' Handler fr das MouseUp-Event des Controls
    ''' </summary>
    Private Sub ListPaintControl_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        ' ein gegebenenfalls vorhandenes MouseDownItem verliert seinen MouseDown-Status
        SetMouseDown(Nothing)

        ' Item ermitteln, ber dem die Maus sich befindet
        Dim NewSelected As ListItem = Nothing
        For Each Item As ListItem In m_KnownItems.Keys
            If m_KnownItems(Item).Rectangle.Contains(e.Location) Then
                NewSelected = Item
                Exit For
            End If
        Next

        ' nur wenn das neu ermittelte Item existiert und sich die Maus auch vor dem Click ber demselben Item
        ' befand, wird dieses als Selected festgelegt
        If NewSelected IsNot Nothing AndAlso NewSelected Is m_HooveredItem Then
            m_Items.SelectedItem = NewSelected
        End If

    End Sub

    ''' <summary>
    ''' Handler fr das Resize-Event des Controls
    ''' </summary>
    Private Sub ListPaintControl_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        ' Position der ScrollBar neu festlegen
        m_ScrollBar.Location = New Point(Me.Width - m_ScrollBar.Width - 1, 1)

        ' auf ScrollBar-Events vorbergehend verzichten
        m_ScrollBarEventsEnabled = False

        ' Hhe anpassen und daraus resultierende nderungen an ScrollBar und PaintItems feststellen
        ChangeHeight(Me.Height)

        ' GradientBrushes zum Zeichnen der Control-Hintergrunds neu ermitteln
        GetBackgroundBrushes()

        ' Ereignisse der ScrollBar wieder zulassen
        m_ScrollBarEventsEnabled = True

        ' Control neu zeichnen
        Invalidate()
    End Sub


    ''' <summary>
    ''' berechnet abhngige Elemente nach nderung der Hhe des Controls neu
    ''' </summary>
    Private Sub ChangeHeight(ByVal NewHeight As Integer)
        ' nur wenn sich die Hhe des Controls wirklich gendert hat
        If m_Height <> NewHeight Then
            ' aktuelle Hhe merken
            m_Height = NewHeight

            ' Hhe der ScrollBar anpassen
            m_ScrollBar.Height = m_Height - 2

            ' Hhe aller Listenelemente zusammen
            Dim ListHeight As Integer = m_Items.Count * m_ItemHeight
            If ListHeight > m_Height Then
                ' Wenn die Listen-Hhe die des Controls bertrifft, mu die ScrollBar angezeigt werden
                m_ScrollBar.Visible = True
                ' Anzeigebreite der Items neu festlegen
                m_ItemWidth = Me.Width - m_ScrollBar.Width - 2
                ' Maximum der ScrollBar festlegen
                m_ScrollBar.Maximum = ListHeight - 1
                ' LargeChange-Eigenschaft der ScrollBar festlegen
                m_ScrollBar.LargeChange = m_Height
                ' maximal beim manuellen Scrollen erreichbarer Wert fr Value-Eigenschaft der ScrollBar
                Dim MaxScrollableValue As Integer = ListHeight - m_Height

                ' Wert der ScrollBar nach Mglichkeit auf den zuletzt beim manuellen Scrollen
                ' eingestellten Wert festlegen, wenn nicht mglich, den maximal manuell erreichbaren Wert verwenden
                m_ScrollBar.Value = Math.Min(m_LastScrolledValue, MaxScrollableValue)

            Else
                ' Wenn die Listen-Hhe kleiner als die des Controls ist, wird auf die ScrollBar verzichtet
                m_ScrollBar.Visible = False
                ' Anzeigebreite der Items neu festlegen
                m_ItemWidth = Me.Width - 1
                ' Value-Eigenschaft der ScrollBar auf 0 stellen
                m_ScrollBar.Value = 0
            End If

            ' sichtbare Items neu ermitteln
            ReSetVisibleItems()
        End If

    End Sub

    ''' <summary>
    ''' berschreiben der OnPaint-Methode zum Neuzeichnen des Controls
    ''' </summary>
    Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
        MyBase.OnPaint(e)
        ' Referenz zum Graphics-Objekt der PaintEventArgs
        Dim g As Graphics = e.Graphics

        ' Flche mit Hintergrundfarbe fllen
        g.Clear(m_Renderer.ControlBackColor)

        ' Flchen mit Farbgradienten zeichnen
        g.FillRectangle(m_LeftGradientBrush, m_LeftGradientRect)
        g.FillRectangle(m_RightGradientBrush, m_RightGradientRect)

        ' rechte Randbegrenzung zeichnen
        g.DrawLine(m_Renderer.ToolStripBorderPen, Me.Width - 1, 0, Me.Width - 1, Me.Height)

        ' sichtbare Items zeichnen
        For Each Item As ListItem In m_KnownItems.Keys
            m_KnownItems(Item).OnPaint(e)
        Next

        ' untere Randbegrenzung zeichnen
        g.DrawLine(m_Renderer.MenuBorderPen, 0, Me.Height - 1, Me.Width, Me.Height - 1)
    End Sub

    ''' <summary>
    ''' berschreiben der OnMouseWheel-Methode des Controls
    ''' </summary>
    Protected Overrides Sub OnMouseWheel(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseWheel(e)

        ' wenn die ScrollBar sichtbar ist, kann auch mit dem Mausrad gescrollt werden
        If m_ScrollBar.Visible Then
            ' neuen ScrollBar-Wert im Rahmen der erlaubten Grenzen ermitteln
            Dim NewValue As Integer = Math.Max(m_ScrollBar.Minimum, (Math.Min(m_ScrollBar.Maximum - Me.Height, m_ScrollBar.Value - e.Delta)))
            ' neuen Wert an ScrollBar bergeben
            ' (in der Folge sorgen Events der ScrollBar fr Neuzeichnen)
            m_ScrollBar.Value = NewValue
        End If

    End Sub

    ''' <summary>
    ''' berschreiben der OnKeyDown-Methode des Controls
    ''' </summary>
    Protected Overrides Sub OnKeyDown(ByVal e As System.Windows.Forms.KeyEventArgs)
        MyBase.OnKeyDown(e)
        ' Mit dem Up- oder Down-Button soll das selektierte Item gendert werden knnen
        ' Dies geschieht, falls mglich, ber die SelectedIndex-Eigenschaft der Liste
        ' In der Folge sorgt das Changed-Event der Liste fr Neuzeichnen
        If e.KeyCode = Keys.Down Then
            If m_Items.SelectedIndex < m_Items.Count - 1 Then
                m_Items.SelectedIndex += 1
            End If
        ElseIf e.KeyCode = Keys.Up Then
            If m_Items.SelectedIndex > 0 Then
                m_Items.SelectedIndex -= 1
            End If
        End If
    End Sub

    ''' <summary>
    '''  berschreiben der IsInputKey-Methode des Controls
    ''' </summary>
    Protected Overrides Function IsInputKey(ByVal keyData As System.Windows.Forms.Keys) As Boolean
        ' aktiviert die Up- und Down-Taste als akzeptierte Eingabetasten
        If keyData = Keys.Up OrElse keyData = Keys.Down Then
            Return True
        End If
        Return MyBase.IsInputKey(keyData)
    End Function

    ''' <summary>
    ''' Handler fr das SelectedItemChanged-Event der Item-Liste 
    ''' </summary>
    Private Sub m_Items_SelectedItemChanged() Handles m_Items.SelectedItemChanged
        If m_Items.SelectedItem IsNot Nothing Then
            ' wenn ein SelectedItem in der Liste enthalten ist, 
            ' ScrollBar so einstellen, da dieses Item auch sichtbar ist
            EnsureVisible(m_Items.SelectedItem)
            ' letzten Value-Wert der ScrollBar merken
            m_LastScrolledValue = m_ScrollBar.Value
        End If
        ' als angezeigtes SelectedItem angezeigtes PaintItem aktualisieren
        SetSelected(m_Items.SelectedItem)
    End Sub

    ''' <summary>
    ''' sorgt dafr, dass gegebenenfalls der Wert der ScrollBar so gendert wird, da  
    ''' ein ausgewhltes Item sichtbar wird
    ''' </summary>
    ''' <param name="Item">ausgewhltes Item</param>
    Private Sub EnsureVisible(ByVal Item As ListItem)
        ' nur wenn die ScrollBar wirklich angezeigt wird
        If m_ScrollBar.Visible Then
            ' Basis ist der aktuelle Wert der ScrollBar
            Dim NewValue As Integer = m_ScrollBar.Value
            ' Anzeigeposition des sichtbar zu machenden Items
            Dim Top As Integer = m_Items.IndexOf(Item) * m_ItemHeight

            If m_ScrollBar.Value > Top Then
                ' wenn der aktuelle Wert der ScrollBar zu gro ist, wird als neuer Wert die Position
                ' des sichtbar zu machenden Items vorgesehen
                NewValue = Top
            ElseIf m_ScrollBar.Value + m_Height < Top + m_ItemHeight Then
                ' wenn das Item ber die untere Grenze dieses Controls hinaus ragt, 
                ' neue Position so ermitteln, da die untere Grenze des Items mit diesem Control abschliet 
                NewValue = Top + m_ItemHeight - m_Height
            End If

            ' nur wenn der neu ermittelte ScrollBar-Wert sich vom aktuellen unterscheidet 
            If NewValue <> m_ScrollBar.Value Then
                ' auf ScrollBar-Events vorbergehend verzichten
                m_ScrollBarEventsEnabled = False
                ' neuen Wert fr die ScrollBar bernehmen
                m_ScrollBar.Value = NewValue
                ' sichtbare Items und Anzeigepositionen neu ermitteln
                ReSetVisibleItems()
                ' Control neu zeichnen
                Invalidate()

                ' ScrollBar-Events wieder zulassen
                m_ScrollBarEventsEnabled = True
            End If
        End If
    End Sub

    ''' <summary>
    ''' Handler fr das VisibleChanged-Event der ScrollBar
    ''' </summary>
    Private Sub m_ScrollBar_VisibleChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles m_ScrollBar.VisibleChanged
        ' neue Item-Breite ermitteln
        If m_ScrollBar.Visible Then
            m_ItemWidth = Me.Width - m_ScrollBar.Width - 2
        Else
            m_ItemWidth = Me.Width - 1
        End If
        ' GradientBrushes zum Zeichnen des Control-Hintergrunds neu ermitteln
        GetBackgroundBrushes()
    End Sub

#End Region

End Class
